from datetime import datetime
import re

from target_gcs.sinks import GCSSink
from target_gcs.target import TargetGCS


def build_sink(config={}):
    default_config = {"bucket_name": "test-bucket", "credentials_file": "./creds.json"}
    config.update(default_config)
    print(232222, config)
    return GCSSink(
        TargetGCS(config=config), "my_stream", {"properties": {}}, key_properties=config
    )


def test_extraction_timestamp_is_unix_time():
    subject = build_sink()
    assert re.match(r"my_stream_\d+.jsonl", subject.key_name)


def test_key_name_includes_prefix_when_provided():
    subject = build_sink({"key_prefix": "asdf"})
    assert re.match(r"asdf/my_stream", subject.key_name)


def test_key_name_does_not_start_with_slash():
    subject = build_sink({"key_prefix": "/asdf"})
    assert not subject.key_name.startswith("/")


def test_key_name_includes_stream_name_when_naming_convention_not_provided():
    subject = build_sink({"key_naming_convention": "asdf.txt"})
    assert "asdf.txt" == subject.key_name


def test_key_name_includes_stream_name_if_stream_token_used():
    subject = build_sink({"key_naming_convention": "___{stream}___.txt"})
    assert "___my_stream___.txt" == subject.key_name


def test_key_name_includes_default_date_format_if_date_token_used():
    date_format = "%Y-%m-%d"
    subject = build_sink({"key_naming_convention": "file/{date}.txt"})
    assert f"file/{datetime.today().strftime(date_format)}.txt" == subject.key_name


def test_key_name_includes_date_format_if_date_token_used_and_date_format_provided():
    date_format = "%m %d, %Y"
    subject = build_sink(
        {"key_naming_convention": "file/{date}.txt", "date_format": date_format}
    )
    assert f"file/{datetime.today().strftime(date_format)}.txt" == subject.key_name


def test_key_name_includes_timestamp_if_timestamp_token_used():
    subject = build_sink({"key_naming_convention": "file/{timestamp}.txt"})
    print(subject.key_name)
    assert re.match(r"file/\d+.txt", subject.key_name)



"""GCS target sink class, which handles writing streams.
Opens an I/O stream pointed to a blob in the GCS bucket
As messages are received from the source, they are streamed to the bucket. As the max batch size is reached, the queue is emptied and written to the I/O streams

"""
from collections import defaultdict
from datetime import datetime
import json
import time

from google.cloud.storage import Client
from singer_sdk.sinks import RecordSink
from smart_open import open


class GCSSink(RecordSink):
    """GCS target sink class."""

    max_size = 1000  # Max records to write in one batch

    def __init__(self, target, stream_name, schema, key_properties):
        super().__init__(
            target=target,
            stream_name=stream_name,
            schema=schema,
            key_properties=key_properties,
        )
        self._gcs_write_handle = None
        self.key_name = self.build_key_name(self.config)

    def build_key_name(self, config):
        extraction_timestamp = round(time.time())
        key_name = self.config.get(
            "key_naming_convention",
            f"{self.stream_name}_{extraction_timestamp}.{self.output_format}",
        )
        key_name = f'{self.config.get("key_prefix", "")}/{key_name}'.replace("//", "/")
        if key_name.startswith("/"):
            key_name = key_name[1:]
        date = datetime.today().strftime(self.config.get("date_format", "%Y-%m-%d"))
        key_name = key_name.format_map(
            defaultdict(
                str, stream=self.stream_name, date=date, timestamp=extraction_timestamp
            )
        )
        return key_name

    @property
    def gcs_write_handle(self):
        """opens a stream for writing to the target cloud object"""
        if not self._gcs_write_handle:
            credentials_path = self.config.get("credentials_file")
            self._gcs_write_handle = open(
                f'gs://{self.config.get("bucket_name")}/{self.key_name}',
                "wb",
                transport_params=dict(
                    client=Client.from_service_account_json(credentials_path)
                ),
            )
        return self._gcs_write_handle

    @property
    def output_format(self):
        """In the future maybe we will support more formats"""
        return "jsonl"

    def process_record(self, record: dict, context: dict) -> None:
        """Process the record.

        Developers may optionally read or write additional markers within the
        passed `context` dict from the current batch.
        """
        self.gcs_write_handle.write(
            f"{json.dumps(record, default=str)}\n".encode("utf-8")
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['target_gcs', 'target_gcs.tests']

package_data = \
{'': ['*']}

install_requires = \
['google-api-python-client>=2.49.0,<3.0.0',
 'google-cloud-storage>=2.3.0,<3.0.0',
 'pytest-watch>=4.2.0,<5.0.0',
 'requests>=2.25.1,<3.0.0',
 'singer-sdk>=0.5.0,<0.6.0',
 'smart-open[gcs]>=6.0.0,<7.0.0']

entry_points = \
{'console_scripts': ['target-gcs = target_gcs.target:TargetGCS.cli']}

setup_kwargs = {
    'name': 'target-gcs',
    'version': '1.0.1',
    'description': '`target-gcs` is a Singer target for GCS, built with the Meltano SDK for Singer Targets.',
    'long_description': None,
    'author': 'Datateer Ops',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""GCS target class."""

from singer_sdk.target_base import Target
from singer_sdk import typing as th

from target_gcs.sinks import (
    GCSSink,
)


class TargetGCS(Target):
    """Sample target for GCS."""

    name = "target-gcs"
    config_jsonschema = th.PropertiesList(
        th.Property("credentials_file", th.StringType, required=True),
        th.Property("bucket_name", th.StringType, required=True),
        th.Property("key_prefix", th.StringType, required=False),
        th.Property("key_naming_convention", th.StringType, required=False),
    ).to_dict()
    default_sink_class = GCSSink
